from autoarray.plot import mat_objs


class LightProfileCentreScatterer(mat_objs.Scatterer):
    def __init__(self, size=None, marker=None, colors=None, from_subplot_config=False):

        super(LightProfileCentreScatterer, self).__init__(
            size=size,
            marker=marker,
            colors=colors,
            section="light_profile_centres",
            from_subplot_config=from_subplot_config,
        )

    @classmethod
    def sub(cls, size=None, marker=None, colors=None):
        return LightProfileCentreScatterer(
            size=size, marker=marker, colors=colors, from_subplot_config=True
        )


class MassProfileCentreScatterer(mat_objs.Scatterer):
    def __init__(self, size=None, marker=None, colors=None, from_subplot_config=False):

        super(MassProfileCentreScatterer, self).__init__(
            size=size,
            marker=marker,
            colors=colors,
            section="mass_profile_centres",
            from_subplot_config=from_subplot_config,
        )

    @classmethod
    def sub(cls, size=None, marker=None, colors=None):
        return MassProfileCentreScatterer(
            size=size, marker=marker, colors=colors, from_subplot_config=True
        )


class MultipleImagesScatterer(mat_objs.Scatterer):
    def __init__(self, size=None, marker=None, colors=None, from_subplot_config=False):

        super(MultipleImagesScatterer, self).__init__(
            size=size,
            marker=marker,
            colors=colors,
            section="multiple_images",
            from_subplot_config=from_subplot_config,
        )

    @classmethod
    def sub(cls, size=None, marker=None, colors=None):
        return MultipleImagesScatterer(
            size=size, marker=marker, colors=colors, from_subplot_config=True
        )


class CriticalCurvesLiner(mat_objs.Liner):
    def __init__(
        self,
        width=None,
        style=None,
        colors=None,
        pointsize=None,
        from_subplot_config=False,
    ):

        super(CriticalCurvesLiner, self).__init__(
            width=width,
            style=style,
            colors=colors,
            pointsize=pointsize,
            section="critical_curves",
            from_subplot_config=from_subplot_config,
        )

    @classmethod
    def sub(cls, width=None, style=None, colors=None, pointsize=None):
        return CriticalCurvesLiner(
            width=width,
            style=style,
            colors=colors,
            pointsize=pointsize,
            from_subplot_config=True,
        )


class CausticsLiner(mat_objs.Liner):
    def __init__(
        self,
        width=None,
        style=None,
        colors=None,
        pointsize=None,
        from_subplot_config=False,
    ):

        super(CausticsLiner, self).__init__(
            width=width,
            style=style,
            colors=colors,
            pointsize=pointsize,
            section="caustics",
            from_subplot_config=from_subplot_config,
        )

    @classmethod
    def sub(cls, width=None, style=None, colors=None, pointsize=None):
        return CausticsLiner(
            width=width,
            style=style,
            colors=colors,
            pointsize=pointsize,
            from_subplot_config=True,
        )



import autoarray as aa
import autoastro as aast

import time

# Although we could test_autoarray the image without using an image (e.aast.lp. by just making a grid), we have chosen to
# set this test_autoarray up using an image and mask. This gives run-time numbers that can be easily related to an actual lens
# analysis

shape_2d = (301, 301)
pixel_scales = 0.05
sub_size = 4
radius = 3.0

print("sub grid size = " + str(sub_size))
print("circular mask radius = " + str(radius) + "\n")

mask = aa.Mask.circular(
    shape_2d=shape_2d, pixel_scales=pixel_scales, sub_size=sub_size, radius=radius
)

grid = aa.Grid.from_mask(mask=mask)

print("Number of grid points = " + str(grid.sub_shape_1d) + "\n")

### EllipticalGaussian ###

light_profile = aast.lp.EllipticalGaussian(
    centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, sigma=1.0
)

start = time.time()
light_profile.profile_image_from_grid(grid=grid)
diff = time.time() - start
print("EllipticalGaussian time = {}".format(diff))

### SphericalGaussian ###

light_profile = aast.lp.SphericalGaussian(centre=(0.0, 0.0), sigma=1.0)

start = time.time()
light_profile.profile_image_from_grid(grid=grid)
diff = time.time() - start
print("SphericalGaussian time = {}".format(diff))

### EllipticalExponential ###

profile = aast.lp.EllipticalExponential(
    centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, intensity=1.0, effective_radius=1.0
)

start = time.time()
profile.profile_image_from_grid(grid=grid)
diff = time.time() - start
print("EllipticalExponential time = {}".format(diff))

### SphericalExponential ###

profile = aast.lp.SphericalExponential(
    centre=(0.0, 0.0), intensity=1.0, effective_radius=1.0
)

start = time.time()
profile.profile_image_from_grid(grid=grid)
diff = time.time() - start
print("SphericalExponential time = {}".format(diff))

### EllipticalDevVaucouleurs ###

profile = aast.lp.EllipticalDevVaucouleurs(
    centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, intensity=1.0, effective_radius=1.0
)

start = time.time()
profile.profile_image_from_grid(grid=grid)
diff = time.time() - start
print("EllipticalDevVaucouleurs time = {}".format(diff))

### SphericalDevVaucouleurs ###

profile = aast.lp.SphericalDevVaucouleurs(
    centre=(0.0, 0.0), intensity=1.0, effective_radius=1.0
)

start = time.time()
profile.profile_image_from_grid(grid=grid)
diff = time.time() - start
print("SphericalDevVaucouleurs time = {}".format(diff))

### EllipticalSersic ###

light_profile = aast.lp.EllipticalSersic(
    centre=(0.0, 0.0),
    axis_ratio=0.8,
    phi=45.0,
    intensity=1.0,
    effective_radius=1.0,
    sersic_index=2.5,
)

start = time.time()
light_profile.profile_image_from_grid(grid=grid)
diff = time.time() - start
print("EllipticalSersic time = {}".format(diff))

### SphericalSersic ###

light_profile = aast.lp.SphericalSersic(
    centre=(0.0, 0.0), intensity=1.0, effective_radius=1.0, sersic_index=2.5
)

start = time.time()
light_profile.profile_image_from_grid(grid=grid)
diff = time.time() - start
print("SphericalSersic time = {}".format(diff))

### EllipticalCoreSersic ###

light_profile = aast.lp.EllipticalCoreSersic(
    centre=(0.0, 0.0),
    axis_ratio=0.8,
    phi=45.0,
    intensity=1.0,
    effective_radius=1.0,
    sersic_index=2.5,
    radius_break=0.01,
    intensity_break=0.05,
    gamma=0.25,
    alpha=3.0,
)

start = time.time()
light_profile.profile_image_from_grid(grid=grid)
diff = time.time() - start
print("EllipticalCoreSersic time = {}".format(diff))

### SphericalCoreSersic ###

light_profile = aast.lp.SphericalCoreSersic(
    centre=(0.0, 0.0),
    intensity=1.0,
    effective_radius=1.0,
    sersic_index=2.5,
    radius_break=0.01,
    intensity_break=0.05,
    gamma=0.25,
    alpha=3.0,
)

start = time.time()
light_profile.profile_image_from_grid(grid=grid)
diff = time.time() - start
print("SphericalCoreSersic time = {}".format(diff))

print()



import autoarray as aa
import autoastro.plot as aplt
import os

import pytest

from os import path

directory = path.dirname(path.realpath(__file__))


@pytest.fixture(name="galaxy_fit_plotter_path")
def make_galaxy_fit_plotter_setup():
    return "{}/files/plots/galaxy_fitting/".format(
        os.path.dirname(os.path.realpath(__file__))
    )


@pytest.fixture(autouse=True)
def set_config_path():
    aa.conf.instance = aa.conf.Config(
        path.join(directory, "files/plotter"), path.join(directory, "output")
    )


def test__fit_sub_plot__all_types_of_galaxy_fit(
    gal_fit_7x7_image,
    gal_fit_7x7_convergence,
    gal_fit_7x7_potential,
    gal_fit_7x7_deflections_y,
    gal_fit_7x7_deflections_x,
    positions_7x7,
    include_all,
    plot_patch,
    galaxy_fit_plotter_path,
):
    aplt.FitGalaxy.subplot_fit_galaxy(
        fit=gal_fit_7x7_image,
        positions=positions_7x7,
        include=include_all,
        sub_plotter=aplt.SubPlotter(
            output=aplt.Output(path=galaxy_fit_plotter_path, format="png")
        ),
    )

    assert galaxy_fit_plotter_path + "subplot_fit_galaxy.png" in plot_patch.paths

    aplt.FitGalaxy.subplot_fit_galaxy(
        fit=gal_fit_7x7_convergence,
        positions=positions_7x7,
        include=include_all,
        sub_plotter=aplt.SubPlotter(
            output=aplt.Output(path=galaxy_fit_plotter_path, format="png")
        ),
    )

    assert galaxy_fit_plotter_path + "subplot_fit_galaxy.png" in plot_patch.paths

    aplt.FitGalaxy.subplot_fit_galaxy(
        fit=gal_fit_7x7_potential,
        positions=positions_7x7,
        include=include_all,
        sub_plotter=aplt.SubPlotter(
            output=aplt.Output(path=galaxy_fit_plotter_path, format="png")
        ),
    )

    assert galaxy_fit_plotter_path + "subplot_fit_galaxy.png" in plot_patch.paths

    aplt.FitGalaxy.subplot_fit_galaxy(
        fit=gal_fit_7x7_deflections_y,
        positions=positions_7x7,
        include=include_all,
        sub_plotter=aplt.SubPlotter(
            output=aplt.Output(path=galaxy_fit_plotter_path, format="png")
        ),
    )

    assert galaxy_fit_plotter_path + "subplot_fit_galaxy.png" in plot_patch.paths

    aplt.FitGalaxy.subplot_fit_galaxy(
        fit=gal_fit_7x7_deflections_x,
        positions=positions_7x7,
        include=include_all,
        sub_plotter=aplt.SubPlotter(
            output=aplt.Output(path=galaxy_fit_plotter_path, format="png")
        ),
    )

    assert galaxy_fit_plotter_path + "subplot_fit_galaxy.png" in plot_patch.paths



import autoastro as aast
from astropy import cosmology as cosmo

import pytest

planck = cosmo.Planck15


class TestCosmology:
    def test__arcsec_to_kpc_conversion(self):

        arcsec_per_kpc = aast.util.cosmology.arcsec_per_kpc_from_redshift_and_cosmology(
            redshift=0.1, cosmology=planck
        )

        assert arcsec_per_kpc == pytest.approx(0.525060, 1e-5)

        kpc_per_arcsec = aast.util.cosmology.kpc_per_arcsec_from_redshift_and_cosmology(
            redshift=0.1, cosmology=planck
        )

        assert kpc_per_arcsec == pytest.approx(1.904544, 1e-5)

        arcsec_per_kpc = aast.util.cosmology.arcsec_per_kpc_from_redshift_and_cosmology(
            redshift=1.0, cosmology=planck
        )

        assert arcsec_per_kpc == pytest.approx(0.1214785, 1e-5)

        kpc_per_arcsec = aast.util.cosmology.kpc_per_arcsec_from_redshift_and_cosmology(
            redshift=1.0, cosmology=planck
        )

        assert kpc_per_arcsec == pytest.approx(8.231907, 1e-5)

    def test__angular_diameter_distances(self):

        angular_diameter_distance_to_earth_kpc = aast.util.cosmology.angular_diameter_distance_to_earth_from_redshift_and_cosmology(
            redshift=0.1, cosmology=planck, unit_length="kpc"
        )

        assert angular_diameter_distance_to_earth_kpc.unit == "kpc"
        assert angular_diameter_distance_to_earth_kpc == pytest.approx(392840, 1e-5)

        angular_diameter_distance_to_earth_arcsec = aast.util.cosmology.angular_diameter_distance_to_earth_from_redshift_and_cosmology(
            redshift=0.1, cosmology=planck, unit_length="arcsec"
        )

        arcsec_per_kpc = aast.util.cosmology.arcsec_per_kpc_from_redshift_and_cosmology(
            redshift=0.1, cosmology=planck
        )

        assert angular_diameter_distance_to_earth_arcsec.unit == "arcsec"
        assert arcsec_per_kpc * angular_diameter_distance_to_earth_kpc == pytest.approx(
            angular_diameter_distance_to_earth_arcsec, 1e-5
        )

        angular_diameter_distance_between_redshifts = aast.util.cosmology.angular_diameter_distance_between_redshifts_from_redshifts_and_cosmlology(
            redshift_0=0.1, redshift_1=1.0, cosmology=planck, unit_length="kpc"
        )

        assert angular_diameter_distance_between_redshifts.unit == "kpc"
        assert angular_diameter_distance_between_redshifts == pytest.approx(
            1481890.4, 1e-5
        )

    def test__critical_surface_mass_densities(self):

        critical_surface_density = aast.util.cosmology.critical_surface_density_between_redshifts_from_redshifts_and_cosmology(
            redshift_0=0.1, redshift_1=1.0, cosmology=planck, unit_length="kpc"
        )

        assert critical_surface_density.unit == "solMass / kpc^2"
        assert critical_surface_density == pytest.approx(4.85e9, 1e-2)

        critical_surface_density = aast.util.cosmology.critical_surface_density_between_redshifts_from_redshifts_and_cosmology(
            redshift_0=0.1,
            redshift_1=1.0,
            cosmology=planck,
            unit_mass="solMass",
            unit_length="arcsec",
        )

        assert critical_surface_density.unit == "solMass / arcsec^2"
        assert critical_surface_density == pytest.approx(17593241668, 1e-2)

    def test__cosmic_average_density(self):

        cosmic_average_density = aast.util.cosmology.cosmic_average_density_from_redshift_and_cosmology(
            redshift=0.6, cosmology=planck, unit_mass="solMass", unit_length="kpc"
        )

        assert cosmic_average_density == pytest.approx(249.20874, 1.0e-4)

        cosmic_average_density = aast.util.cosmology.cosmic_average_density_from_redshift_and_cosmology(
            redshift=0.6, cosmology=planck, unit_mass="solMass", unit_length="arcsec"
        )

        assert cosmic_average_density == pytest.approx(81280.09116133313, 1.0e-4)



import autofit as af
import autoastro as aast

import pytest


@pytest.fixture(autouse=True)
def reset_config():
    """
    Use configuration from the default path. You may want to change this to set a specific path.
    """
    af.conf.instance = af.conf.default


class TestCase:
    def test_integration(self):
        # Create a mapper. This can be used to convert values output by a non linear optimiser into class instances.
        mapper = af.ModelMapper()

        # Create a model_galaxy prior for the source model_galaxy. Here we are describing only the light profile of
        # the source model_galaxy which comprises an elliptical exponential and elliptical sersic light profile.
        source_galaxy_prior = aast.GalaxyModel(
            redshift=aast.Redshift,
            light_profile_one=aast.lp.EllipticalExponential,
            light_profile_2=aast.lp.EllipticalSersic,
        )

        # Create a model_galaxy prior for the source model_galaxy. Here we are describing both the light and mass
        # profiles. We've also stipulated that the centres of any galaxies generated using the model_galaxy prior
        # should match.
        lens_galaxy_prior = aast.GalaxyModel(
            redshift=aast.Redshift,
            light_profile=aast.lp.EllipticalExponential,
            mass_profile=aast.mp.EllipticalExponential,
            align_centres=True,
        )

        mapper.source_galaxy = source_galaxy_prior
        mapper.lens_galaxy = lens_galaxy_prior

        # Create a model instance. All the instances of the profile classes are created here. Normally we would do this
        # using the output of a non linear search but in this case we are using the median values from the priors.
        instance = mapper.instance_from_prior_medians()

        # Recover model_galaxy instances. We can pass the model instance to model_galaxy priors to recover a fully
        # constructed model_galaxy
        source_galaxy = instance.source_galaxy
        lens_galaxy = instance.lens_galaxy

        # Let's just check that worked
        assert len(source_galaxy.light_profiles) == 2
        assert len(source_galaxy.mass_profiles) == 0

        assert len(lens_galaxy.light_profiles) == 1
        assert len(lens_galaxy.mass_profiles) == 1

        assert source_galaxy.redshift == 1.5
        assert lens_galaxy.redshift == 1.5



import autoarray as aa
import autoastro as aast

import time


# Although we could test_autoarray the deflection angles without using an image (e.aast.mp. by just making a grid), we have chosen to
# set this test_autoarray up using an image and mask. This gives run-time numbers that can be easily related to an actual lens
# analysis

shape_2d = (301, 301)
pixel_scales = 0.05
sub_size = 2
radius = 4.0

print("sub grid size = " + str(sub_size))
print("circular mask radius = " + str(radius) + "\n")


mask = aa.Mask.circular(
    shape_2d=shape_2d, pixel_scales=pixel_scales, sub_size=sub_size, radius=radius
)

grid = aa.Grid.from_mask(mask=mask)

print("Number of points = " + str(grid.sub_shape_1d) + "\n")

### EllipticalGaussian (Integrate) ###

mass_profile = aast.mp.EllipticalGaussian(
    centre=(0.0, 0.0),
    axis_ratio=0.6,
    phi=0.0,
    intensity=1.0,
    sigma=10.0,
    mass_to_light_ratio=1.0,
)

start = time.time()
mass_profile.deflections_from_grid(grid=grid)
diff = time.time() - start
print("EllipticalGaussian time = {}".format(diff))



import autoastro as aast

from test_autoarray.unit.conftest import *

directory = path.dirname(path.realpath(__file__))


@pytest.fixture(autouse=True)
def set_config_path():
    conf.instance = conf.Config(
        path.join(directory, "config"), path.join(directory, "output")
    )


#
# MODEL #
#

# PROFILES #


@pytest.fixture(name="lp_0")
def make_lp_0():
    # noinspection PyTypeChecker
    return aast.lp.SphericalSersic(
        intensity=1.0, effective_radius=2.0, sersic_index=2.0
    )


@pytest.fixture(name="lp_1")
def make_lp_1():
    # noinspection PyTypeChecker
    return aast.lp.SphericalSersic(
        intensity=2.0, effective_radius=2.0, sersic_index=2.0
    )


@pytest.fixture(name="mp_0")
def make_mp_0():
    # noinspection PyTypeChecker
    return aast.mp.SphericalIsothermal(einstein_radius=1.0)


@pytest.fixture(name="mp_1")
def make_mp_1():
    # noinspection PyTypeChecker
    return aast.mp.SphericalIsothermal(einstein_radius=2.0)


@pytest.fixture(name="lmp_0")
def make_lmp_0():
    return aast.lmp.EllipticalSersicRadialGradient()


@pytest.fixture(name="dmp_0")
def make_dmp_0():
    # noinspection PyTypeChecker
    return aast.mp.SphericalNFW(kappa_s=1.0)


@pytest.fixture(name="dmp_1")
def make_dmp_1():
    # noinspection PyTypeChecker
    return aast.mp.SphericalNFW(kappa_s=2.0)


@pytest.fixture(name="smp_0")
def make_smp_0():
    # noinspection PyTypeChecker
    return aast.lmp.EllipticalSersic(intensity=1.0, mass_to_light_ratio=1.0)


@pytest.fixture(name="smp_1")
def make_smp_1():
    # noinspection PyTypeChecker
    return aast.lmp.EllipticalSersic(intensity=2.0, mass_to_light_ratio=2.0)


# GALAXY #


@pytest.fixture(name="gal_x1_lp")
def make_gal_x1_lp(lp_0):
    return aast.Galaxy(redshift=0.5, light_profile_0=lp_0)


@pytest.fixture(name="gal_x2_lp")
def make_gal_x2_lp(lp_0, lp_1):
    return aast.Galaxy(redshift=0.5, light_profile_0=lp_0, light_profile_1=lp_1)


@pytest.fixture(name="gal_x1_mp")
def make_gal_x1_mp(mp_0):
    return aast.Galaxy(redshift=0.5, mass_profile_0=mp_0)


@pytest.fixture(name="gal_x2_mp")
def make_gal_x2_mp(mp_0, mp_1):
    return aast.Galaxy(redshift=0.5, mass_profile_0=mp_0, mass_profile_1=mp_1)


@pytest.fixture(name="gal_x1_lp_x1_mp")
def make_gal_x1_lp_x1_mp(lp_0, mp_0):
    return aast.Galaxy(redshift=0.5, light_profile_0=lp_0, mass_profile_0=mp_0)


@pytest.fixture(name="hyper_galaxy")
def make_hyper_galaxy():
    return aast.HyperGalaxy(noise_factor=1.0, noise_power=1.0, contribution_factor=1.0)


# GALAXY DATA #


@pytest.fixture(name="gal_data_7x7")
def make_gal_data_7x7(image_7x7, noise_map_7x7):
    return aast.GalaxyData(
        image=image_7x7, noise_map=noise_map_7x7, pixel_scales=image_7x7.pixel_scales
    )


@pytest.fixture(name="gal_fit_data_7x7_image")
def make_gal_fit_data_7x7_image(gal_data_7x7, sub_mask_7x7):
    return aast.MaskedGalaxyDataset(
        galaxy_data=gal_data_7x7, mask=sub_mask_7x7, use_image=True
    )


@pytest.fixture(name="gal_fit_data_7x7_convergence")
def make_gal_fit_data_7x7_convergence(gal_data_7x7, sub_mask_7x7):
    return aast.MaskedGalaxyDataset(
        galaxy_data=gal_data_7x7, mask=sub_mask_7x7, use_convergence=True
    )


@pytest.fixture(name="gal_fit_data_7x7_potential")
def make_gal_fit_data_7x7_potential(gal_data_7x7, sub_mask_7x7):
    return aast.MaskedGalaxyDataset(
        galaxy_data=gal_data_7x7, mask=sub_mask_7x7, use_potential=True
    )


@pytest.fixture(name="gal_fit_data_7x7_deflections_y")
def make_gal_fit_data_7x7_deflections_y(gal_data_7x7, sub_mask_7x7):
    return aast.MaskedGalaxyDataset(
        galaxy_data=gal_data_7x7, mask=sub_mask_7x7, use_deflections_y=True
    )


@pytest.fixture(name="gal_fit_data_7x7_deflections_x")
def make_gal_fit_data_7x7_deflections_x(gal_data_7x7, sub_mask_7x7):
    return aast.MaskedGalaxyDataset(
        galaxy_data=gal_data_7x7, mask=sub_mask_7x7, use_deflections_x=True
    )


# GALAXY FIT #


@pytest.fixture(name="gal_fit_7x7_image")
def make_gal_fit_7x7_image(gal_fit_data_7x7_image, gal_x1_lp):
    return aast.FitGalaxy(
        masked_galaxy_dataset=gal_fit_data_7x7_image, model_galaxies=[gal_x1_lp]
    )


@pytest.fixture(name="gal_fit_7x7_convergence")
def make_gal_fit_7x7_convergence(gal_fit_data_7x7_convergence, gal_x1_mp):
    return aast.FitGalaxy(
        masked_galaxy_dataset=gal_fit_data_7x7_convergence, model_galaxies=[gal_x1_mp]
    )


@pytest.fixture(name="gal_fit_7x7_potential")
def make_gal_fit_7x7_potential(gal_fit_data_7x7_potential, gal_x1_mp):
    return aast.FitGalaxy(
        masked_galaxy_dataset=gal_fit_data_7x7_potential, model_galaxies=[gal_x1_mp]
    )


@pytest.fixture(name="gal_fit_7x7_deflections_y")
def make_gal_fit_7x7_deflections_y(gal_fit_data_7x7_deflections_y, gal_x1_mp):
    return aast.FitGalaxy(
        masked_galaxy_dataset=gal_fit_data_7x7_deflections_y, model_galaxies=[gal_x1_mp]
    )


@pytest.fixture(name="gal_fit_7x7_deflections_x")
def make_gal_fit_7x7_deflections_x(gal_fit_data_7x7_deflections_x, gal_x1_mp):
    return aast.FitGalaxy(
        masked_galaxy_dataset=gal_fit_data_7x7_deflections_x, model_galaxies=[gal_x1_mp]
    )


from autoastro.plot.lensing_plotters import Include

# PLOTTING #


@pytest.fixture(name="include_all")
def make_include_all():
    return Include(
        origin=True,
        mask=True,
        grid=True,
        border=True,
        positions=True,
        light_profile_centres=True,
        mass_profile_centres=True,
        critical_curves=True,
        caustics=True,
        multiple_images=True,
        inversion_pixelization_grid=True,
        inversion_grid=True,
        inversion_border=True,
        inversion_image_pixelization_grid=True,
        preloaded_critical_curves=[(1.0, 1.0), (2.0, 2.0)],
        preload_caustics=[(1.0, 1.0), (2.0, 2.0)],
    )



from astropy import constants
from autoastro import exc
from autoastro import dimensions as dim
import math


def arcsec_per_kpc_from_redshift_and_cosmology(redshift, cosmology):
    return cosmology.arcsec_per_kpc_proper(z=redshift).value


def kpc_per_arcsec_from_redshift_and_cosmology(redshift, cosmology):
    return 1.0 / cosmology.arcsec_per_kpc_proper(z=redshift).value


def angular_diameter_distance_to_earth_from_redshift_and_cosmology(
    redshift, cosmology, unit_length="kpc"
):

    angular_diameter_distance_kpc = cosmology.angular_diameter_distance(z=redshift).to(
        "kpc"
    )

    if unit_length is not "arcsec":
        angular_diameter_distance = angular_diameter_distance_kpc.to(unit_length)
    else:
        arcsec_per_kpc = arcsec_per_kpc_from_redshift_and_cosmology(
            redshift=redshift, cosmology=cosmology
        )
        angular_diameter_distance = arcsec_per_kpc * angular_diameter_distance_kpc

    return dim.Length(angular_diameter_distance.value, unit_length)


def angular_diameter_distance_between_redshifts_from_redshifts_and_cosmlology(
    redshift_0, redshift_1, cosmology, unit_length="kpc"
):

    if unit_length is "arcsec":
        raise exc.UnitsException(
            "The angular diameter distance between redshifts cannot have unit_label of length"
        )

    angular_diameter_distance_between_redshifts_kpc = cosmology.angular_diameter_distance_z1z2(
        redshift_0, redshift_1
    ).to(
        "kpc"
    )

    if unit_length is not "arcsec":
        angular_diameter_distance = angular_diameter_distance_between_redshifts_kpc.to(
            unit_length
        )

    return dim.Length(angular_diameter_distance.value, unit_length)


def cosmic_average_density_from_redshift_and_cosmology(
    redshift, cosmology, unit_length="arcsec", unit_mass="solMass"
):

    cosmic_average_density_kpc = cosmology.critical_density(z=redshift).to(
        unit_mass + " / kpc^3"
    )

    cosmic_average_density_kpc = dim.MassOverLength3(
        value=cosmic_average_density_kpc.value, unit_length="kpc", unit_mass=unit_mass
    )

    if unit_length is not "arcsec":
        cosmic_average_density = cosmic_average_density_kpc.convert(
            unit_length=unit_length, unit_mass=unit_mass
        )
    else:
        kpc_per_arcsec = kpc_per_arcsec_from_redshift_and_cosmology(
            redshift=redshift, cosmology=cosmology
        )
        cosmic_average_density = cosmic_average_density_kpc.convert(
            unit_length=unit_length, unit_mass=unit_mass, kpc_per_arcsec=kpc_per_arcsec
        )

    return cosmic_average_density


def critical_surface_density_between_redshifts_from_redshifts_and_cosmology(
    redshift_0, redshift_1, cosmology, unit_length="arcsec", unit_mass="solMass"
):

    if unit_mass is "angular":
        return dim.MassOverLength2(
            value=1.0, unit_mass=unit_mass, unit_length=unit_length
        )

    const = constants.c.to("kpc / s") ** 2.0 / (
        4 * math.pi * constants.G.to("kpc3 / (" + unit_mass + " s2)")
    )

    angular_diameter_distance_of_redshift_0_to_earth_kpc = angular_diameter_distance_to_earth_from_redshift_and_cosmology(
        redshift=redshift_0, cosmology=cosmology, unit_length="kpc"
    )

    angular_diameter_distance_of_redshift_1_to_earth_kpc = angular_diameter_distance_to_earth_from_redshift_and_cosmology(
        redshift=redshift_1, cosmology=cosmology, unit_length="kpc"
    )

    angular_diameter_distance_between_redshifts_kpc = angular_diameter_distance_between_redshifts_from_redshifts_and_cosmlology(
        redshift_0=redshift_0,
        redshift_1=redshift_1,
        cosmology=cosmology,
        unit_length="kpc",
    )

    critical_surface_density_kpc = (
        const
        * angular_diameter_distance_of_redshift_1_to_earth_kpc
        / (
            angular_diameter_distance_between_redshifts_kpc
            * angular_diameter_distance_of_redshift_0_to_earth_kpc
        )
    )

    critical_surface_density_kpc = dim.MassOverLength2(
        value=critical_surface_density_kpc.value, unit_mass=unit_mass, unit_length="kpc"
    )

    if unit_length not in "arcsec":

        critical_surface_density = critical_surface_density_kpc.convert(
            unit_length=unit_length, unit_mass=unit_mass
        )

    elif unit_length in "arcsec":
        kpc_per_arcsec = kpc_per_arcsec_from_redshift_and_cosmology(
            redshift=redshift_0, cosmology=cosmology
        )
        critical_surface_density = critical_surface_density_kpc.convert(
            unit_mass=unit_mass, unit_length=unit_length, kpc_per_arcsec=kpc_per_arcsec
        )

    return critical_surface_density


def scaling_factor_between_redshifts_from_redshifts_and_cosmology(
    redshift_0, redshift_1, redshift_final, cosmology
):

    angular_diameter_distance_between_redshifts_0_and_1 = (
        cosmology.angular_diameter_distance_z1z2(z1=redshift_0, z2=redshift_1)
        .to("kpc")
        .value
    )

    angular_diameter_distance_to_redshift_final = (
        cosmology.angular_diameter_distance(z=redshift_final).to("kpc").value
    )

    angular_diameter_distance_of_redshift_1_to_earth = (
        cosmology.angular_diameter_distance(z=redshift_1).to("kpc").value
    )

    angular_diameter_distance_between_redshift_1_and_final = (
        cosmology.angular_diameter_distance_z1z2(z1=redshift_0, z2=redshift_final)
        .to("kpc")
        .value
    )

    return (
        angular_diameter_distance_between_redshifts_0_and_1
        * angular_diameter_distance_to_redshift_final
    ) / (
        angular_diameter_distance_of_redshift_1_to_earth
        * angular_diameter_distance_between_redshift_1_and_final
    )



import autofit as af
import autoastro as aast
import numpy as np

import pytest


@pytest.fixture(autouse=True)
def reset_config():
    """
    Use configuration from the default path. You may want to change this to set a specific path.
    """
    af.conf.instance = af.conf.default


class TestHyperImageSky:
    def test__scale_sky_in_image__increases_all_image_values(self):

        image = np.array([1.0, 2.0, 3.0])

        hyper_sky = aast.hyper_data.HyperImageSky(sky_scale=10.0)

        scaled_image = hyper_sky.hyper_image_from_image(image=image)

        assert (scaled_image == np.array([11.0, 12.0, 13.0])).all()


class TestHyperNoiseMapBackground:
    def test__scaled_background_noise__adds_to_input_noise(self):

        noise_map = np.array([1.0, 2.0, 3.0])

        hyper_background_noise_map = aast.hyper_data.HyperBackgroundNoise(
            noise_scale=2.0
        )

        hyper_noise_map = hyper_background_noise_map.hyper_noise_map_from_noise_map(
            noise_map=noise_map
        )

        assert (hyper_noise_map == np.array([3.0, 4.0, 5.0])).all()

        hyper_noise_map_background = aast.hyper_data.HyperBackgroundNoise(
            noise_scale=3.0
        )

        scaled_noise = hyper_noise_map_background.hyper_noise_map_from_noise_map(
            noise_map=noise_map
        )

        assert (scaled_noise == np.array([4.0, 5.0, 6.0])).all()



from autoarray.plot.mat_objs import (
    Units,
    Figure,
    ColorMap,
    ColorBar,
    Ticks,
    Labels,
    Legend,
    Output,
    OriginScatterer,
    MaskScatterer,
    BorderScatterer,
    GridScatterer,
    PositionsScatterer,
    IndexScatterer,
    PixelizationGridScatterer,
    Liner,
    VoronoiDrawer,
)

from autoarray.plot import imaging_plots as Imaging
from autoarray.plot import interferometer_plots as Interferometer
from autoarray.plot import fit_imaging_plots as FitImaging
from autoarray.plot import fit_interferometer_plots as FitInterferometer
from autoarray.plot import mapper_plots as Mapper
from autoarray.plot import inversion_plots as Inversion

from autoastro.plot.lensing_mat_objs import (
    LightProfileCentreScatterer,
    MassProfileCentreScatterer,
    MultipleImagesScatterer,
    CriticalCurvesLiner,
    CausticsLiner,
)

from autoastro.plot.lensing_plotters import Plotter, SubPlotter, Include

from autoastro.plot.lensing_plotters import plot_array as Array
from autoastro.plot.lensing_plotters import plot_grid as Grid
from autoastro.plot.lensing_plotters import plot_line as Line
from autoastro.plot.lensing_plotters import plot_mapper_obj as MapperObj

from autoastro.plot import light_profile_plots as LightProfile
from autoastro.plot import mass_profile_plots as MassProfile
from autoastro.plot import galaxy_plots as Galaxy
from autoastro.plot import fit_galaxy_plots as FitGalaxy
